#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import math
import numpy as np

sys.path.append('../')

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
 
from redcp import * # import data types

import pickle #pickle for reading .pkl

from utils import *

from Planner import Planner

# Results dir
RESULT_DIR = 'results'
ROBOT_SIZE = 0.5
EPS = 0.001

# - Obstacle growing ----------------------------------------------------------
def evaluate_obstacle_growing(planner, scenarios, reference, radius):
    for idx in np.arange(0,len(scenarios)): #evaluate individual scenarios obstacle growing
        gridmap, start, goal = scenarios[idx]
        ref_grid, ref_path = reference[idx]
        
        print("Obstacle growing test (Scenario %d)" % idx)
        gridmap_processed = planner.grow_obstacles(gridmap, radius)

        #plot the gridmap
        fig = plot_map(plt, gridmap, 'Obstacle growing - Input')
        image_file = RESULT_DIR + '/obstacle_growing-' + str(idx) + '-map.png'
        save_image(fig, 'Save obstacle growing input map to ' + image_file, image_file)

        #plot the solution
        fig = plot_map(plt, gridmap_processed, 'Obstacle growing - Solution')
        image_file = RESULT_DIR + '/obstacle_growing-' + str(idx) + '-solution.png'
        save_image(fig, 'Save obstacle growing solution to ' + image_file, image_file)

        #plot the reference map
        fig = plot_map(plt, ref_grid, 'Obstacle growing - Reference solution')
        image_file = RESULT_DIR + '/obstacle_growing-' + str(idx) + '-reference.png'
        save_image(fig, 'Save obstacle growing reference solution to ' + image_file, image_file)

        #plot the difference
        diff = ref_grid.clone()
        data = diff.data.reshape(diff.height, diff.width)
        data_g = gridmap_processed.data.reshape(gridmap_processed.height, gridmap_processed.width)
        data = np.abs(data - data_g)
        diff.data = data.flatten()
        fig = plot_map(plt, diff, 'Obstacle growing - Solution Difference from Reference Solution')
        image_file = RESULT_DIR + '/obstacle_growing-' + str(idx) +'-difference.png'
        save_image(fig, 'Save obstacle growing solution difference to ' + image_file, image_file)
    return

###########################################################
# - PATH PLANNING
def evaluate_path_planning(planner, scenarios, reference, radius):
    for idx in np.arange(0,len(scenarios)):
        gridmap, start, goal = scenarios[idx]
        ref_grid, ref_path = reference[idx]
        plt.close('all')

        gridmap_processed = planner.grow_obstacles(gridmap, radius)

        print("Path planning (Scenario %d)" % idx)
        path = planner.plan_path(gridmap_processed, start, goal)

        print("Path simplification (Scenario %d)" % idx)
        path_simple = planner.simplify_path(gridmap_processed, path)
        
        fig = plot_paths(plt, path, path_simple, gridmap_processed, "Path planning - Solution")
        image_file = RESULT_DIR + '/path_planning-' + str(idx) + '-solution.png'
        save_image(fig, 'Save path planning solution to ' + image_file, image_file)

        fig = plot_paths(plt, ref_path, None, gridmap_processed, "Path planning - Reference Solution")
        image_file = RESULT_DIR + '/path_planning-' + str(idx) + '-reference.png'
        save_image(fig, 'Save path planning reference solution to ' + image_file, image_file)
        
        path_length = get_path_length(path)
        ref_length = get_path_length(ref_path)
        print('Solution path length %.2f' % path_length)
        print('Reference path length %.2f' % ref_length)

        if path_length == 0:
            if ref_length == 0:
                print("Path simplification correct\n")
            else:   
                print("ERROR: Non-existent path has been returned!\n")
        else:
            if ref_length == 0:
                print("ERROR: Non-existent path has been returned!\n")
            else:
                if (np.abs(ref_path.poses[0].position.x - path.poses[0].position.x) > EPS or
                    np.abs(ref_path.poses[0].position.y - path.poses[0].position.y) > EPS or 
                    np.abs(ref_path.poses[-1].position.x - path.poses[-1].position.x) > EPS or
                    np.abs(ref_path.poses[-1].position.y - path.poses[-1].position.y) > EPS):

                    if (np.abs(ref_path.poses[0].position.x - path.poses[-1].position.x) <= EPS and 
                        np.abs(ref_path.poses[0].position.y - path.poses[-1].position.y) <= EPS and 
                        np.abs(ref_path.poses[-1].position.x - path.poses[0].position.x) <= EPS and
                        np.abs(ref_path.poses[-1].position.y - path.poses[0].position.y) <= EPS):
                        print("ERROR: Path in reverse order\n")
                    else:
                        print("ERROR: Missing start or goal pose in path\n")

                    if path_simple is not None:
                        if len(path_simple.poses) > 2 and len(path_simple.poses) < len(path.poses):
                            print("Path simplification correct\n")
                        else:
                            print("ERROR: Returned simplified path does contain only two poses or is of similar length as the original path")
                else:
                    if not check_path_validity(path, gridmap, EPS):
                        print('ERROR: The student path is not valid.')
                    elif path_length <= ref_length:
                        print('Optimal path found')
                    else: 
                        print('WARN: Optimal path not found. If you have implemented non-optimal planner the solution need to be further checked\n If use 8-neighborhood optimal planner, consider that there is most likely a mistake in your solution\n')

                    if path_simple is not None:
                        if len(path_simple.poses) > 2 and len(path_simple.poses) < len(path.poses):
                            print("Path simplification correct\n")
                        else:
                            print("ERROR: Simplified path does contain only two poses or is of similar length as the original path\n")
    return
        
if __name__=="__main__":
    if not os.path.exists(RESULT_DIR):
        os.mkdir(RESULT_DIR)

    ###################################################
    # Evaluate the planner 
    planner = Planner()

    #load the dataset
    scenarios = pickle.load( open( "data/scenarios.pkl", "rb" ) )
    reference = pickle.load( open( "data/plan_results.pkl", "rb" ) )
    print('Loaded scenarios dataset and reference')

    evaluate_obstacle_growing(planner, scenarios, reference, ROBOT_SIZE)
    #evaluate_obstacle_growing(planner, scenarios, reference, ROBOT_SIZE - 0.05)
    evaluate_path_planning(planner, scenarios, reference, ROBOT_SIZE)

