
from ExplorerBase import *

class Explorer(ExplorerBase):

    STOP_BUFFER_STEPS = 5

    def __init__(self):
        ExplorerBase.__init__(self)

    def mapping(self):
        """ Mapping thread for fusing the laser scans into the grid map
        """
        counter = 1
        stop = False
        while stop is not None:
            ## fuse the laser scan   
            #print("[Mapping] Fuse laser scan %4d" % counter)
            self.mutex_gridmap.acquire() ## call in mutex, because scan must be fused one at a time
            laser_scan = self.robot.laser_scan_
            odometry = self.robot.odometry_
            counter += 1
            self.gridmap = self.mapper.fuse_laser_scan(self.gridmap, laser_scan, odometry) ## Use grid node to handle fusing the laser scan
            self.mutex_gridmap.release()
            time.sleep(self.SLEEP_MAPPING) ## Plot the map
            self.mutex_stop.acquire()
            stop = self.stop
            self.mutex_stop.release()
        print('[Mapping] Terminate')
    
    def planning(self):
        """ Planning thread that takes the constructed gridmap, find frontiers, and select the next goal with the navigation path  
        """
        time.sleep(2) ## little delay before mapping is done
        stop = False
        stop_counter = self.STOP_BUFFER_STEPS
        counter = 1
        while not stop:
            time.sleep(self.SLEEP_PLANNING)
            if self.gridmap is None or self.gridmap.data is None: continue
            self.mutex_gridmap.acquire()
            nav_gridmap = self.gridmap.clone() # make copy of the gridmap
            start = self.robot.odometry_.pose
            self.mutex_gridmap.release()

            #if not self.do_plan: continue
            print("[Planning] Enter planning loop %4d" % counter)
            counter += 1

            # TODO grow map but keep the unknown cells 
            grid = nav_gridmap.data.reshape((nav_gridmap.height, nav_gridmap.width))

            frontiers = self.frontiers_extractor.find_free_edge_frontiers_representatives(nav_gridmap)

            print("[Planning] No. of detected frontiers %d" % len(frontiers))
            frontiers = [f for f in frontiers if f.dist(start) >= DELTA_DISTANCE]
            print("[Planning] No. of frontiers after DELTA_DISTANCE reduction %d" % len(frontiers))

            #TODO cleanup the current robot position in the map as it is freespace regardless missing laser data

            min_dist = sys.float_info.max
            goal_pose = None
            for pose in frontiers: # determination of the closest representatives using Euclidean distance
                pth = self.planner.plan_path(nav_gridmap, start, pose)

                #TODO select best waypoint based on path length (among grown obstacles)
                dist = start.dist(pose)  
                if dist < min_dist:
                    min_dist = dist
                    goal_pose = pose
            path = simplified_path = None
            if goal_pose is not None:
                path = self.planner.plan_path(nav_gridmap, start, goal_pose)
            if path is not None:
                print('[Planning] New path planned (pose) %.2f %.2f -> %.2f %.2f' % (start.position.x, start.position.y, goal_pose.position.x, goal_pose.position.y))
                simplified_path = self.planner.simplify_path(nav_gridmap, path) ## path simplification
                print('[Planning] Simplified len %f' % (simplified_path.len(),))
            else:
                print('[Planning] Path not found')

            self.mutex_planning.acquire()
            self.frontiers = frontiers
            self.nav_gridmap = nav_gridmap
            self.planned_path = path
            self.simplified_planned_path = simplified_path
            self.mutex_planning.release()

            self.mutex_path_following.acquire()
            if path is not None and len(path.poses) > 1:
                self.waypoints = path.poses[1:]
                self.new_plan = True
                print('[Planning] Set new plan in waypoints')
            self.mutex_path_following.release()

            self.mutex_stop.acquire()
            if frontiers is None or len(frontiers) == 0 and stop_counter <= 0:
                self.stop = True
                print('[Planning] No more frontiers, terminate exploration!')
            elif frontiers is None or len(frontiers) == 0:
                print('[Planning] No more frontiers, decrement stop counter %d' % stop_counter)
                stop_counter -= 1
            else:
                stop_counter = self.STOP_BUFFER_STEPS

            stop = self.stop
            self.mutex_stop.release()
            print('[Planning] tick done')
        print('[Planning] terminated!')
 
    def path_following(self):
        """Path following thread that assigns new goals to the robot navigation thread
        """ 
        waypoints = [ #assign initial goals for navigation
                Pose(Vector3(3.5, 3.5, 0), Quaternion(1, 0, 0, 0)), 
                Pose(Vector3(0.0, 0.0, 0), Quaternion(1, 0, 0, 0)), 
        ]
        stop = False  
        next_wp = True # if true fetch new waypoint
        while not stop:
            #print('[Path following] tick')
            if next_wp:
                if waypoints is not None and len(waypoints) > 1:
                    print('[Path following] New navigation goal')
                    self.robot.goto_reactive(waypoints.pop(0))
                else:
                    self.robot.move(None) #stop the robot and reset its navigation_goal
                next_wp = False
            goal = self.robot.get_navigation_goal() # Call getter for mutex 
            if goal is not None:
                odom = self.robot.odometry_.pose #check the robot distance to goal
                self.path.poses.append(odom)  
                odom.position.z = 0 #compensate for the height of the robot as we are interested only in achieved planar distance 
                dist = goal.dist(odom) #calculate the distance
                print("[Path following] Distance to the actual goal %.2f" % dist)
                if dist < WAYPOINT_REACH_DISTANCE:
                    print("[Path following] Goal reached, ask for waypoint")
                    next_wp = True
            else:
                print("[Path following] Null goal, ask for waypoint")
                next_wp = True #try to fetch next wp

            time.sleep(self.SLEEP_PATH_FOLLOW) # sleep for a while

            self.mutex_path_following.acquire() # acquire lock for update shared data
            if self.new_plan: # check if a new plan has been issued
                print('[Path following] New waypoints to be followed received')
                self.new_plan = False
                next_wp = True
                waypoints = self.waypoints
            self.mutex_path_following.release()

            self.mutex_stop.acquire()
            stop = self.stop
            self.mutex_stop.release()
        print('[Path following] Terminate')

